#include "net.h"
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include "compat.h"
#include "util.h"

/** Nastavi non-blocking na file deskriptor */
void set_nonblock(int fd) {
 int fl;
 chkerror("set_nonblock:fcntl()",fl=fcntl(fd,F_GETFL));
 chkerror("set_nonblock:fcntl()",fcntl(fd,F_SETFL,fl|O_NONBLOCK));
// setsockopt(s,SOL_SOCKET,SO_NONBLOCK,&noblock, sizeof(int));
}

/** stejne jako send, az na to, ze se send_all pokusi prostlacit vzdy celou zpravu
 (pokud send vrati ze byla poslana pouze cast zpravy, send_all pokracuje pustenim send na zbytek)
*/
int send_all(int sock,const void *buf,size_t buf_bytes,int flags) {
 int sbytes;
 size_t remaining=buf_bytes;
 char *b=(char *)buf;
 do {
  sbytes=send(sock,b,remaining,flags);
  if (sbytes==-1) return -1;//chyba pri posilani - vratim tedy chybu
  remaining-=sbytes;
  b+=sbytes;
  if (sbytes==0) { /// nepodarilo se nic poslat, ale neni chyba - takze druhy konec aasi zavrel spojeni
   return buf_bytes-remaining;//vratim pocet bajtu, co jsem skutecne poslal  
  }
 } while (remaining);//opakuj dokud je co posilat
 return buf_bytes;
}



/** posle retezec + CRLF na socket */
void send_line(int s,char *str) {
 int l=strlen(str);
 char buf[1024];//buffer
 char *rbuf;
 if (l>sizeof(buf)-2) { //alokuj buffer jen pokud je potreba - 1kb vetsinou staci
  rbuf=malloc(l+2);
 } else rbuf=buf;
 memcpy(rbuf,str,l);
 rbuf[l]='\r';
 rbuf[l+1]='\n';
 send_all(s,rbuf,l+2,MSG_NOSIGNAL);
 if (rbuf!=buf) free(rbuf);
}

/** jako send_line(), ale s formatovanym retezcem jako u printf */
void send_linef(int sock,char *fmt,...){
 va_list argptr;
 char s[1024];
 va_start(argptr,fmt);	
 _vsnprintf(s,sizeof(s),fmt,argptr);
 va_end(argptr);
 send_line(sock,s);
}

/** zacne naslouchat na portu [port], vrati socket */
int net_listen(int port) {
 int s;
 struct sockaddr_in addr;

 addr.sin_family=AF_INET;
 addr.sin_port=htons(port);
 addr.sin_addr.s_addr=htonl(INADDR_ANY);

 s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
 chkerror("socket()",s);
 chkerror("bind()",bind(s,(struct sockaddr *)(&addr),sizeof(addr)));
 chkerror("listen()",listen(s,5));

 set_nonblock(s);
 return s;
}

/** prevede port na cislo */
int port_from_string(const char *port) {
 struct servent *s;
 s=getservbyname(port,"tcp");
 if (s) {
//  debug_printf("found port -> %s = %d\n",port,ntohs(s->s_port));
  return ntohs(s->s_port);
 }
 return atoi(port);
}

/** pro dany host vrati jeho IPV4 adresu */
long resolve(const char *name) {
 struct hostent *ph;
 if (!name || !*name) return INADDR_NONE;
// if (isaddrp(name)) return ntohl(inet_addr(name));
 ph=gethostbyname2(name,AF_INET);
 if (!ph) return INADDR_NONE;
 if (ph->h_length!=4 || ph->h_addrtype!=AF_INET) return INADDR_NONE;
 return ntohl(*((long *)(ph->h_addr)));
}


/** otevre konexi na dany host a port, vrati socket nebo -1 pokud chyba */
int open_connection(const char *host,int port) {
 struct sockaddr_in inx;
 int sock;
 long ip;
 int i;

// debug_printf ("open connection: %s:%d\n",host,port);

 sock=socket(AF_INET,SOCK_STREAM,IPPROTO_IP);
 memset(&inx,0,sizeof(inx));
 if (sock==-1) return -1;
 ip=resolve(host);
 if (ip==INADDR_NONE) return -1;
// debug_printf("Connecting to %ld.%ld.%ld.%ld\n",ipv4(ip));
 inx.sin_family=AF_INET;
 inx.sin_port=htons(port);
 inx.sin_addr.s_addr=htonl(ip);
 i=connect(sock,(struct sockaddr*)&inx,sizeof(inx));
 if (i==-1) {
  close(sock);
  return -1;
 }
 return sock;
}

/** priconnecti se na adresu ve tvaru (hostname|adresa)(:port)?
 def_port se pouzije, pokud port neni zadan*/
int connect_string(const char *host,int def_port) {
 char *host2;
 char *pos;
 int s;
 host2=strdup(host);
 pos=strrchr(host2,':');
 if (pos) {
  *pos=0;
  pos++;
  def_port=port_from_string(pos);
 }
 //printf("Connectiong to: %s:%d",host2,def_port);
 s=open_connection(host2,def_port);
 free(host2);
 return s;
}

/** nacte radku ze socketu, zbyle znaky (za koncem radky)
 vrati do bufferu, pokud je radka jiz v bufferu, nic nenacita.
 Pri uzavreni spojeni vraci NULL*/
char *socket_getline(charbuffer *b, int sock) {
 char buf[1024];//buffer na data
 char *line;//nactena radka
 int rb;//nactene bajty
 line=buf_getline(b);
 if (line) return line;
 while((rb=recv(sock,buf,sizeof(buf),0))) {
  buf_add(b,buf,rb);
  line=buf_getline(b);
  if (line) return line;
 }
 return NULL;
}
