#include "config.h"
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <pthread.h>
#include <syslog.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "util.h"
#include "cfg_file.h"
#include "net.h"
#include "buffer.h"
#include "compat.h"

#define ipv4(x) ((x)>>24)&0xff,((x)>>16)&0xff,((x)>>8)&0xff,(x)&0xff //IP adresa jako 4 slozky

/** zkoukne commandline parametry a pak podle nich nahraje konfiguracni soubor */
void init_config(int argc,char **argv) {
 int i;
 struct sigaction act;
 for(i=1;i<argc;i++) {
  if (eq(argv[i],"-v")) {
   printf (PROGRAM_NAME " " PROGRAM_VERSION "\nCompiled: " __DATE__ " " __TIME__ "\n");
   exit(0);
  }
  if (eq(argv[i],"-c") && (i<argc-1)) {
   configfile=argv[i+1];
   i++;
  } else fatalf("Invalid commandline switch: %s",argv[i]);
 }
 if (!configfile) configfile=DEFAULT_CONFIG_FILE;
 active=load_config(configfile);
 //Ted to nastavi signal handler
 act.sa_handler=sighup_handler;
 sigemptyset(&act.sa_mask);
 act.sa_flags=0;
 chkerror("initconfig:sigaction()",sigaction(SIGHUP,&act,NULL));
}

char valid_char[256];//platne znaky v emailu

/** inicializuje tabulku valid_char */
void make_vch_table() {
 int i;
 static char *vchars="._-";
 for (i=0;i<=255;i++) valid_char[i]=0;
 for (i='A';i<='Z';i++) valid_char[i]=1;
 for (i='a';i<='z';i++) valid_char[i]=1;
 for (i='0';i<='9';i++) valid_char[i]=1;
 for (i=0;i<strlen(vchars);i++) valid_char[(int)(vchars[i])]=1;
}

/** Escape na tecky v smtp komunikaci
 cnt==-1: z .. na jednom radku udela . 
 cnt==+1: z . na jednom radku udela .. */
void message_dots(charbuffer *b,int cnt) {
 int i,len;
 int shift=0;
 char *buf=b->buf;
 len=b->size;
 if (cnt==-1) { //odstran
  if (len<4) return;//"\n..\n" alespon tam byt musi
  for (i=0;i<len-3;i++) {
   if (newline(buf[0]) && buf[1]=='.' && buf[2]=='.' && newline(buf[3])) {
    debug_printf("dvojtecka!\n");
    shift++; //dalsi dvojtecka
   }
   buf[3-shift]=buf[3];//kopiruje znaky vzad
   buf++;
  }
  b->size-=shift;
 } else {
  if (len<3) return;//"\n.\n" alespon tam byt musi
  for (i=0;i<b->size-2;i++) {
   if (newline(buf[0]) && buf[1]=='.' && newline(buf[3])) {
    debug_printf("jednotecka!\n");
    buf_realloc(b,b->size+1);
    shift++; //jen jedna tecka
    memmove(buf+2,buf+1,b->size-i-1);//posun zpravu, dvojujic tak tecku (aby to klient nebral jako konec zpravy)
    b->size++;
   }
   buf++;
  }
 }
}

/** vraci true, pokud podretezec retezce string delky strlen je email fe formatu user@domena */
int is_canonical_email(const char *string,int s_len) {
 int i,z;
 z=0;
 for(i=0;i<s_len;i++) {
  if (string[i]=='@') z++;
  else if (!valid_char[(int)(string[i])]) {
   if (!(z==0 && string[i]=='+')) {//nektere znaky jsou platne jen v username
    z=-1;
    break;
   }
  }
 }
 return z==1;
}

/** rozparsuje email z formatu typu "neco" <user@domena>
a podobnych na normalizovany tvar user@domena */
char *parse_email(const char *email) {
 char *tmp,*out;
 int l2;
 int l=strlen(email);
 while(l && is_space(email[l-1])) l--;
 //je to primo format user@domena ?
 if (is_canonical_email(email,l)) return strdup(email);//ANO!
 //Format "neco" <user@domena>
 //eventuelne =?charset?q?neco= <user@domena>
 //eventuelne jeste horsi haraburdi <user@domena>
 if (email[l-1]=='>') {
  tmp=strrchr(email,'<');
  if (tmp) { //uzavreno ve spicatych zavorkach
   tmp++;
   l2=l-(tmp-email)-1;
   if (is_canonical_email(tmp,l2)) {//a je tam email
    out=malloc(l2+1);
    memcpy(out,tmp,l2);
    out[l2]=0;
    return out;
   }
  }
 }
 //Format user@domena (neco)
 if (email[l-1]==')') {
  tmp=strrchr(email,'(');
  if (tmp) { //ano, je to ono
   l2=(tmp-email)-1;
   if (is_canonical_email(email,l2)) {//a je tam email
    out=malloc(l2+1);
    memcpy(out,email,l2);
    out[l2]=0;
    return out;
   }   
  }
 }
 return strdup(email);//buhvi co to je, vratim v puvodni podobe
}

typedef int (threadfunc)(int param,config *cfg);//funkce threadu

typedef struct { //zaznam pro jeden thread
 threadfunc *func;	//funkce pro thread
 int sock;		//socket pro thread
} thread_rec;

/** wrapper kolem funkce socketu. Funkce (vnitrni) po sobe socket nezvira, to dela tato funkce */
void net_thread(thread_rec *t) {
 int return_socket;
 config *curr;
 curr=lock_config(NULL); 
 debug_printf("config : %x\n",(int)curr);
 return_socket=t->func(t->sock,curr);
 debug_printf("return_socket : %d\n",return_socket);
 if (return_socket!=-1) close(return_socket);
 unlock_config(curr);
 close(t->sock);
 free(t);
 return;
}

/** spusti sitovy thread s danou funkcin a socketem */
void net_thread_run(thread_rec t) {
 thread_rec *dup_t;
 pthread_t thread;
 dup_t=malloc1(sizeof(thread_rec));
 memcpy(dup_t,&t,sizeof(thread_rec));
 //Nastartuje net_thread(dup_t); v novem threadu
 pthread_create(&thread,NULL,(void*)&net_thread,(void*)dup_t);
}

pthread_mutex_t whitelist_lock;//lock na zmenu whitelistu

/** inicializuje logovani hlasek */
void log_init() {
 openlog("sp_proxy",LOG_CONS,LOG_DAEMON);
}

/** zaloguje hlasky. Typicky do syslogu, ale pri debugovani i na STDOUT */
void log_message(int priority,const char *err) {
 syslog(priority,"%s",err);
 debug_printf("[SYSLOG/%d]%s\n",priority,err);
}

/** jako log_message(), ale s formatovanym retezcem jako u printf */
void log_messagef(int priority, const char *fmt,...){
 va_list argptr;
 char s[512];
 va_start(argptr,fmt);	
 _vsnprintf(s,sizeof(s),fmt,argptr);
 va_end(argptr);
 log_message(priority,s);
}

/** posle do logu chybovou hlasku podle kodu v ERRNO */
void log_error() {
 log_message(LOG_ERR,strerror(errno));
}

/** zapise adresu do whitelistu */
void add_to_whitelist(config *cfg,const char *line) {
 int w;//whitelist file
 char b=0;
 if (empty(cfg->whitelistfile)) {
  debug_printf("No whitelist to add to\n");
  return;//neni kam pridavat
 }
 pthread_mutex_lock(&whitelist_lock);
 if (!tree_search(cfg->whitelist,line)) { //este tam neni
  tree_add(&cfg->whitelist,line,(void *)1);//je jedno co se tam vlozi jako value
  w=open(cfg->whitelistfile,O_CREAT | O_RDWR);
  if (w!=-1) {//soubor jsme otevreli
   lseek(w,-1,SEEK_END);
   read(w,&b,1);
   if (b!='\n') {//soubor nekonci na newline
    b='\n';
    write(w,&b,1);//zapiseme newline
   }
   write(w,line,strlen(line));//zapiseme adresu
   write(w,&b,1);//zapiseme newline
   close(w);
  } else { //zaloguj chybu
   log_error();
  }
 }
 pthread_mutex_unlock(&whitelist_lock);
}

/** otestuje jestli vstup input zacina na cmd a vrati
 odkaz na prvni nemezerovy znak za prikazem cmd, nebo NULL, pokud to neni spravny prikaz */
const char *get_cmd_param(const char *cmd,const char *input) {
 if (!str_beginwithi(cmd,input)) return NULL;//neni to prikaz cmd
 input+=strlen(cmd);
 while (is_space(input[0])) input++;
 return input;
}

/** ZKONTROLUJE jestli radka neobsahuje ^RCPT TO:(.*)$
a pokud ano, prida \1 do whitelistu */
void line_checkwhitelist(config *cfg,char *line) {
 const char *param;
 param=get_cmd_param("rcpt to:",line);
 if (param) add_to_whitelist(cfg,param);
}

/**
SMTP proxy - start double pipe, pass input/output unchanged, if input is
^RCPT TO:(.*)$
then add $1 to whitelist (maybe: only if 250 response code received)
if input is DATA, then wait for .
*/
//	    "RCPT TO:$email <CRLF>
int smtp_handler(int s,config *cfg) {
 int ss=-1,client_eof=0,server_eof=0;
 char buf[1024];//buffer na data
 char *line;//nactena radka
 int rb;//nactene bajty v bufferu
 int rc;
 fd_set fds;
 charbuffer b;
 debug_printf("starting smtp handler");
 buf_init(&b,512);
 ss=connect_string(cfg->remotesmtp,DEFAULT_SMTP_PORT);
 if (ss==-1) {
  debug_printf("SMTP: Connection to %s refused\n",cfg->remotesmtp);
  send_line(s,"421 Connection refused");
  return -1;
 }
 FD_ZERO(&fds);
 FD_SET(s,&fds);
 FD_SET(ss,&fds);
 while ((rc=select(max(s,ss)+1,&fds,NULL,NULL,NULL))) {
  if (rc==-1 && errno==EINTR) continue; //zas ten pitomej signal
  if (FD_ISSET(s,&fds)) {
   rb=recv(s,buf,sizeof(buf),MSG_NOSIGNAL);//nacteme
   if (rb==0 || rb==-1) { //EOF nebo chyba spojeni od klienta
    debug_printf("client EOF\n");
    client_eof=1;
    shutdown(ss,1);//uz na server nelze nic posilat
   } else send_all(ss,buf,rb,MSG_NOSIGNAL);//posleme
   //a ted se jukneme do bufferu jestli tam neni nejaky email, co chceme dat do whitelistu
   buf_add(&b,buf,rb);
   while ((line=buf_getline(&b))) {
    line_checkwhitelist(cfg,line);
    free(line);
   }
   debug_printf("data from client (%d)\n",rb);
  }
  if (FD_ISSET(ss,&fds)) { //jen preposilame, co rika smtp server nas celkem nezajima
   rb=recv(ss,buf,sizeof(buf),MSG_NOSIGNAL);//nacteme
   if (rb==0 || rb==-1) { //EOF od serveru || chyba spojeni
    debug_printf("server EOF\n");
    server_eof=1;
    shutdown(s,1);//uz na klient nelze nic posilat
   } else send_all(s,buf,rb,MSG_NOSIGNAL);//posleme
   debug_printf("data from server (%d)\n",rb);
  }
  debug_printf("done\n");
  //zase vrati sety do puvodniho stavu
  FD_ZERO(&fds);
  if (!client_eof) FD_SET(s,&fds);
  if (!server_eof) FD_SET(ss,&fds);
  if (server_eof && client_eof) break;//konec spojeni
 }
 debug_printf("bailout - end of connection\n");
 while ((line=buf_getline(&b))) {
  line_checkwhitelist(cfg,line);
  free(line);
 }
 buf_free(&b);
 return ss;
}

/** vraci 1 pokud je email v seznamu list, jinak 0 */
int email_in_list(tree *list,const char *email) {
 char *real_email,*domain;
 int result;
 real_email=parse_email(email);//user@domena
 domain=strrchr(real_email,'@');//@domena
 if (!domain) domain="@";//Bez domeny (specialni wildcard)
 result=tree_search(list,real_email) || tree_search(list,domain);
 free(real_email);
 return result;
}

typedef enum {
 pt_filter,
 pt_exec,
 pt_subjectmark,
 pt_classify,
} programtype;


/** spusti program a k nemu oboustrannou pipe - vrati handle na zapis do programu (h_in) a handle na cteni programu (h_out)
Vrati tez pid child procesu (aby se dalo cekat pres waitpid, az vypise vsechen vystup) */
void run_pipe(const char *prog, int *h_in,int *h_out,pid_t *pid) {
 int p_in[2],p_out[2];//read/write pipe file handles
 int err_code;
 pid_t p_pid;
 #ifdef NOT_USE_SH
 int i;
 char *prog_prog,*prog_x,*prog_param;
 char **prog_v;
 int prog_par;
 #endif
 chkerror("run_pipe:pipe(in)",pipe(p_in));
 chkerror("run_pipe:pipe(out)",pipe(p_out));
 debug_printf("will fork soon\n");
 p_pid=chkerror("run_pipe:fork()",fork());
 switch(p_pid) {
  case 0: //Child
   debug_printf("child\n");
   close(p_in[1]);//stdin writing end
   close(p_out[0]);//stdout reading end
   close(0);//STDIN
   dup(p_in[0]);
   close(1);//STDOUT
   dup(p_out[1]);
   close(p_in[0]);//presmerovan stdin z roury na realny stdin
   close(p_out[1]);//presmerovan stdout z roury na realny stdout
   //child si ted prehodil rouru na stdin/stdout a udela exec
   //ZDE zustal otevren z roury pouze STDIN a STDOUT
   #ifdef NOT_USE_SH
   //TODO: jinak, spustit bez shellu
   while (is_space(prog[0])) prog++;//mezery na zacatku nas nezajimaj (stejne by tam nemely byt)
   prog_x=prog_prog=strdup(prog);
   prog_par=0;
   while (prog_x[0]) {
    if (prog_x[0]=='"') {
     prog_x++;
     while (prog_x[0] && (prog_x[0]!='"')) prog_x++; //chybi li uvozovka, dojede az na konec
     if (prog_x[0]) prog_x++;//preskoc uvozovku
    } else if (prog_x[0]=='\'') {
     prog_x++;
     while (prog_x[0] && (prog_x[0]!='\'')) prog_x++; //chybi li uvozovka, dojede az na konec
     if (prog_x[0]) prog_x++;//preskoc uvozovku
    } else {
     while (prog_x[0] && !is_space(prog_x[0])) prog_x++;
    }
    while (prog_x[0] && is_space(prog_x[0])) prog_x++; //preskocit mezery
    prog_par++;
   }
   prog_v=malloc((prog_par+1)*sizeof(char*));
   prog_x=prog_prog;
   for (i=0;i<prog_par;i++) {//doplni NULL znaky
    prog_v[i]=prog_x;
    if (prog_x[0]=='\'') {
     prog_v[i]++;
     prog_x++;
     while (prog_x[0] && (prog_x[0]!='\'')) prog_x++; //chybi li uvozovka, dojede az na konec
     if (prog_x[0]) {prog_x[0]=0;prog_x++;}//preskoc uvozovku
    } else if (prog_x[0]=='"') {
     prog_v[i]++;
     prog_x++;
     while (prog_x[0] && (prog_x[0]!='"')) prog_x++; //chybi li uvozovka, dojede az na konec
     if (prog_x[0]) {prog_x[0]=0;prog_x++;}//preskoc uvozovku
    } else {
     while (prog_x[0] && !is_space(prog_x[0])) prog_x++;
    }
    prog_param=prog_x;
    while (prog_x[0] && is_space(prog_x[0])) prog_x++; //preskocit mezery
    prog_param[0]=0;
    debug_printf("p%d:%s\n",i,prog_v[i]);
   }
   prog_v[prog_par]=NULL;
   err_code=execvp(prog_prog,prog_v);  
   free(prog_v);
   free(prog_prog);
   #else
   err_code=execlp("/bin/sh","/bin/sh","-c",prog,NULL);
   #endif
   if (err_code==-1) { //exec selhal
    debug_printf("exec failed\n");
    exit(255);
   }
   //sem se to stejne nedostane
   exit(EXIT_FAILURE);
   return;
  default: //Parent
   debug_printf("parent\n");
   close(p_in[0]);//stdin reading end
   close(p_out[1]);//stdout writing end
   *h_in=p_in[1];
   *h_out=p_out[0];
   //oba konce roury se predaji ven
   *pid=p_pid;
   return;
 }
}

/** zapise obsah msg do h_in, zaroven vyzvedne obsah z h_out do
 bufferu msg_out, a  po ukonceni procesu pid vrati jeho navratovou hodnotu */
int pipe_filter(charbuffer *msg,charbuffer *msg_out,pid_t pid,int h_in,int h_out) {
 fd_set rs,ws;

 int br=0;//bytes read <- h_out
 int bw=0;//bytes written -> h_in
 int b;
 int input_sent=0;
 int status;//navratovy kod child procesu
 //nastavi noblocking na file descriptory
 set_nonblock(h_in);
 set_nonblock(h_out);

 for(;;) {
  FD_ZERO(&rs);
  FD_SET(h_out,&rs);

  if (!input_sent) {
   FD_ZERO(&ws);
   FD_SET(h_in,&ws);
   b=select(max(h_in,h_out)+1,&rs,&ws,NULL,NULL);
  } else {
   b=select(h_out+1,&rs,NULL,NULL,NULL);//vstup jsme poslali
  }
  if (b==-1 && errno==EINTR) continue; //zas ten pitomej signal
  if (b==-1) goto fail;
  if(!input_sent && FD_ISSET(h_in,&ws)) { //write, dokud je co
   if (bw<msg->size) {//je co psat
    b=write(h_in,msg->buf+bw,msg->size-bw);
    if (b==-1) goto fail;
    bw+=b;
   } else { // uz jsme poslali vsechno, tak zavrem vstup
    debug_printf("will close STDIN...");
    close(h_in);
    debug_printf("closed\n");
    input_sent=1;
   }
  }
  if(FD_ISSET(h_out,&rs)) { //read
   buf_realloc(msg_out,msg_out->size+4096);//zvetsit o 4 kb
   b=read(h_out,msg_out->buf+br,msg_out->a_size-br);
   if (b==-1) goto fail;
   br+=b;
   msg_out->size=br;
   if (b==0) break;//prave jsme nacetli end-of-file
  }
 }
 //Data jsem poslal a proces uz mi taky nic dalsiho poslat nechce
 debug_printf("exit-waiting\n");
 //zavru file descriptory
 close(h_in);
 close(h_out);
 //a ted pockame na exit status procesu       
 if (waitpid(pid,&status,0)!=pid) return -1;//proces asi chcipnul jeste nez jsem se dostal sem
 if (WIFEXITED(status)) return WEXITSTATUS(status);//normalne se ukoncil
 return -1;//proces umrel (asi signal)
fail:
 //neco se pokazilo: vratim original
 buf_realloc(msg_out,msg->size);
 memcpy(msg_out->buf,msg->buf,msg->size);
 close(h_in);
 close(h_out);
 return -1;
}

/** spusti program, msg je buffer, ktery obsahuje vstupni zpravu, pokud pt je
 pt_subjectmark, nebo pt_filter, bude nahrazena vystupni zpravou */
int run_prog(charbuffer *msg,const char *prog,programtype pt) {
 int h_in,h_out;//handle na psani/cteni programu
 pid_t pid;//pid programu
 char *buf;
 int offset;
 charbuffer result;//buffer na vysledek filtrace
 int res_code;
 if (pt==pt_subjectmark) {
  int l;
  debug_printf("-- Mark subject: %s\n",prog);
  buf=msg->buf;
  for(;;) {
   if (str_beginwithi("subject:",buf)) {
    buf+=8;//preskocit "subject:"
    debug_printf("MARK\n");
    while (is_space(buf[0])) buf++;//pereskocit mezery an zacatku
    //Mno a mam misto kam to vlozit
    l=strlen(prog);
    if (msg->size+l>msg->a_size) { //maly buffer, realokuji
     buf_realloc(msg,msg->size+l);
    }
    offset=buf-msg->buf;  
    memmove(buf+l,buf,msg->size-offset);//odsunout data dale
    memmove(buf,prog,l);//vlozit znacku
    msg->size+=l;
    return 0;
   }
   buf=strchr(buf,'\n');//dalsi radka
   if (!buf) break;//posledni radka
   buf++;
   if (buf>msg->buf+msg->size) break;//konec bufferu
   if (*buf=='\r' || *buf=='\n') break; //akorat prazdna radka - skoncily hlavicky
  }
  debug_printf("NO SUBJ\n");
  //zprava neobsauje Subject: - tak ho vlozim
  l=strlen("Subject: \r\n")+strlen(prog);
  buf_realloc(msg,msg->size+l);
  memmove(buf+l,buf,msg->size);//odsunout data dale
  memcpy(buf,"Subject: ",9);//vlozit celou radku
  memcpy(buf+9,prog,strlen(prog));
  memcpy(buf+l-2,"\r\n",2);//vlozit znacku
  msg->size+=l;
  return 0; 
 }
 //spusti se program
 debug_printf("-- exec something: %s\n",prog);
 run_pipe(prog,&h_in,&h_out,&pid);
 debug_printf("-- pipe setup pid=%d\n",pid);
 buf_init(&result,512);
 res_code=pipe_filter(msg,&result,pid,h_in,h_out);
 debug_printf("-- something finished: ret=%d\n",res_code);
 //close(h_in); pipe_filter si je zavre
 //close(h_out); pipe_filter si je zavre
 if (res_code) { //program vratil chybovy kod
  debug_printf("Error (%d) while executing program '%s'",res_code,prog);
  log_messagef(LOG_WARNING,"Error (%d) while executing program '%s'",res_code,prog);
  buf_free(&result);
  return 0;
 }
 debug_printf("-- check result");
 switch(pt) { //program spusten, co ted s vysledkem?
  case pt_filter://filtrace emailu
   if (result.size) {
    buf_free(msg);//originalni zlikviduji
    *msg=result;//a nahradim vysledkem
    return 0;
   } else return 0;//nic nevratil
  case pt_exec: //spusteni programu
   buf_free(&result);//co vyplivl na stdout nas nezajima, originalni zpravu nechame
   return 0;
  case pt_classify: //chceme klasifikaci - vratime hodnotu, original zpravu nechame
   buf=buf_firstline(&result);
   if (!buf) {//nic nevratil! Nebo nevratil celou radku ukoncenou newlinem
    debug_printf("%d bytes\n",result.size);
    log_messagef(LOG_WARNING,"Classificator '%s' have not returned any value!",prog);
    return 0;
   }
   offset=atoi(buf);//prevedeme na cislo
   free(buf);
   buf_free(&result);//co vyplivl na stdout nas nezajima
   debug_printf("\n >>>>> CLASSIFIED: %d <<<<<<\n\n",offset);
   return offset;//vratime hodnotu klasifikace
  default:;//sem to stejne nedojde
 }
 return 0;
}

/** spusti vsechny programy postupne v jejich poradi ze seznamu list
 na zpravu msg, programy dostanou parametry state a pt */
void run_programlist(charbuffer *msg,priority_list *filt,int state,programtype pt) {
 while(filt) {
  if (state==-1 || (state>=filt->min && state<=filt->max)) { //spravna hodnota
   run_prog(msg,filt->prog,pt);
  }
  filt=filt->next;
 }
}

/** spusti na zpravy b_message prefilter, vrati prefiltrovanou zpravu */
charbuffer run_prefilter(charbuffer *b_message,priority_list *filt) {
 charbuffer out;
 out.a_size=out.size=b_message->size;
 out.buf=malloc(out.size);
 memcpy(out.buf,b_message->buf,out.size);
 run_programlist(&out,filt,-1,pt_filter);
 return out;
}
 
/** na zpravu spusti programy postfilter dle nastaveni */
void run_postfilter(charbuffer *b_message,priority_list *filt,int state) {
 run_programlist(b_message,filt,state,pt_filter);
}

/** na zpravu spusti programy postaction dle nastaveni */
void run_postaction(charbuffer *b_message,priority_list *filt,int state) {
 run_programlist(b_message,filt,state,pt_exec);
}

/** na zpravu spusti znackovac Subjectu dle nastaveni */
void run_subjmark(charbuffer *b_message,priority_list *filt,int state) {
 run_programlist(b_message,filt,state,pt_subjectmark);
}

/** na zpravu spusti klasifikator dle nastaveni a vrati hodnotu zpravy */
int run_classificator(charbuffer *b_message,const char *prog) {
 return run_prog(b_message,prog,pt_classify);
}

/** POP3 socket handler */
int pop3_handler(int s,config *cfg) {
 int ss=-1,client_eof=0,server_eof=0;
 int i;
 int rc;
 int is_blacklisted;
 int is_whitelisted;
 int state;//vnistrni stav
 char buf[1024];//buffer na data
 char *line=0,*line_remote=0;//nactena radka
 char *username;//pop3 username
 char *hostname;//pop3 hostname
 const char *param;
 int rb;//nactene bajty v bufferu
 fd_set fds;
 charbuffer b,b_remote,b_message;
 debug_printf("starting pop3 handler");
 buf_init(&b,512);
 buf_init(&b_remote,512);

 send_line(s,"OK sp_proxy pop3 ready");
 nextcommand:
 while ((line=socket_getline(&b,s))) {
  debug_printf("CMD: %s\n",line);
  param=get_cmd_param("quit",line);
  if (param) {//user@host expected
   send_line(s,"+OK bye");
   buf_free(&b);
   buf_free(&b_remote);
   free(line);
   return ss;
  }  
  param=get_cmd_param("user",line);
  if (param) {//user@host expected
   hostname=strrchr(param,'@');
   if (!hostname) {
    send_line(s,"-ERR Send USER in realuser@pop3host format!");
   } else {
    username=strdup(param);
    i=(hostname-param);
    username[i]=0;//"prerizneme" parametr na user a host cast.
    hostname=username+i+1;
    debug_printf("Connecting to %s / %s\n",username,hostname);
    ss=connect_string(hostname,DEFAULT_POP3_PORT);
    if (ss==-1) {
     debug_printf("Connection refused\n");
     send_line(s,"-ERR Connection refused");
     free(line);
     continue;//nacist dalsi prikaz
    }
    line_remote=socket_getline(&b_remote,ss);//nacte greeting message od serveru
    if (str_beginwithi("+OK",line_remote)) { //greeting message je +OK
     send_linef(ss,"USER %s",username);
     free(line_remote);
     line_remote=socket_getline(&b_remote,ss);
     if (str_beginwithi("+OK",line_remote)) { //na user je +OK
      send_line(s,line_remote);//posle odpoved serveru klientovi
      free(line_remote);
      free(line);
      break;//dalsi faze: pripojen na server
     } else {
      send_line(s,line_remote);//posle odpoved serveru klientovi
      free(line_remote);
      free(line);
      continue;//nacist dalsi prikaz
     }
    } else { //pop3 server po pripojeni rek neco jineho nez +OK
     debug_printf("bad response: %s\n",line_remote);
     send_line(s,line_remote);//posleme dal
     free(line_remote);
     free(line);
     continue;//nacist dalsi prikaz
    }
   }
  } else {
   send_line(s,"-ERR Send USER command first!");
  }
  free(line);
 }

 if (!line) { //zavrel pripojeni
  buf_free(&b);
  buf_free(&b_remote);
  return ss;
 }

 debug_printf("Connected to %s / %s\n",username,hostname);
 free(username);//uz ho nepotrebujem
 //TED jsme pripojeni na server, poslali jsme username

 FD_ZERO(&fds);
 FD_SET(s,&fds);
 FD_SET(ss,&fds);
 while ((rc=select(max(s,ss)+1,&fds,NULL,NULL,NULL))) {
  if (rc==-1 && errno==EINTR) continue; //zas ten pitomej signal
  if (FD_ISSET(s,&fds)) {
   rb=recv(s,buf,sizeof(buf),MSG_NOSIGNAL);//nacteme
   if (rb==0 || rb==-1) { //EOF od klienta || chyba spojeni
    debug_printf("client EOF\n");
    client_eof=1;
    shutdown(ss,1);//uz na server nelze nic posilat
   } else {
    buf_add(&b,buf,rb);
    while ((line=buf_getline(&b))) {
     param=get_cmd_param("pass",line);
     if (param) {
      send_line(ss,line);
      free(line);
      line_remote=socket_getline(&b_remote,ss);//nacte odpoved
      send_line(s,line_remote);//odpoved serveru poslu klientovi
      if (!str_beginwithi("+OK",line_remote)) { //na user NENI +OK
       debug_printf("PASS failed!\n");//blby heslo
       free(line_remote);
       send_line(ss,"QUIT");
       shutdown(ss,1);//zavreme spojeni
       close(ss);//odpoved necht si shnije v /dev/null
       ss=-1;
       goto nextcommand;//zpet na zacatek protokolu: ocekavam USER
      }
      free(line_remote);//je to ok, heslo proslo
      continue;//dalsi radka od klienta
     }
     param=get_cmd_param("retr",line);
     if (param) {
      debug_printf("client send RETR %s\n",param);
      send_line(ss,line);//posleme na server
      free(line);
      line_remote=socket_getline(&b_remote,ss);//nacte odpoved
      send_line(s,line_remote);//odpoved serveru poslu klientovi
      debug_printf("- we have response\n");
      if (!str_beginwithi("+OK",line_remote)) { //na retr NENI +OK
       debug_printf("RETR failed!\n");
       free(line_remote);
      } else { //Prijmu zpravu a zpracuju ji
       state=0;//headers
       is_whitelisted=0;
       is_blacklisted=0;
       buf_init(&b_message,4096);
       debug_printf("- we get message\n");
       for(;;) {//nekonecny cyklus
        line_remote=socket_getline(&b_remote,ss);//nacte odpoved
        if (!line_remote) { //spadlo spojeni - koncime
         free(line_remote);
         buf_free(&b);
         buf_free(&b_remote);
         buf_free(&b_message);
         return ss;
        }
        if (eq(line_remote,".")) break;//mam celou message
        buf_addline(&b_message,line_remote);
        if (empty(line_remote)) state=1;//prazdna radka - zacina telo zpravy
        if (state==0) { // prijimam hlavicky
         param=get_cmd_param("from:",line_remote);
         if (param) {
          //TODO: hlavicka muze byt obcas pres vic radek
          if (email_in_list(cfg->whitelist,param)) is_whitelisted=1;
          if (email_in_list(cfg->blacklist,param)) is_blacklisted=1;
         }
        }
        free(line_remote);
       }
       free(line_remote);
       debug_printf("- we have message\n");
       message_dots(&b_message,-1); // vytipat tecky pripadne
       //v b_message je cela zprava
       if (is_whitelisted) state=cfg->whitelist_level;
       else if (is_blacklisted) state=cfg->blacklist_level;
       else { //prozene se to classification (+) a classify-filter
        charbuffer b2;
        debug_printf("- we classify\n");

        b2=run_prefilter(&b_message,cfg->p_prefilt);
        state=run_classificator(&b2,cfg->p_class);
        buf_free(&b2);
       }
       debug_printf("- we run filters (level=%d)\n",state);
       run_postfilter(&b_message,cfg->p_postfilt,state);
       run_postaction(&b_message,cfg->p_postact,state);
       run_subjmark(&b_message,cfg->subject_mark,state);
       message_dots(&b_message,+1); // zdvojit tecky pripadne
       debug_printf("- we send result\n");
       //Mno a ted uz tu zpravu jen poslu
       i=send_all(s,b_message.buf,b_message.size,MSG_NOSIGNAL);
       buf_free(&b_message);
       if (i==-1) { //vypadnout ven, spadlo spojeni
        debug_printf("connection problem\n");
        goto fatal_error;
       }
       send_line(s,".");//a na konci nesmime zapomenout poslat tecku
       debug_printf("- all sent\n");
      }
      continue;
     }
     //zbyle prikazy se jan tak poslou a pak se preposle zpet odpoved (pri dalsim pruchodu selectem)
     send_line(ss,line);
     free(line);
    }
   }
//   debug_printf("data from client (%d)\n",rb);
  }
  if (FD_ISSET(ss,&fds)) {
   //jen preposilame data z pop3 serveru
   //v teto casti preposilame reakci na nejaky neznamy prikaz
   rb=recv(ss,buf,sizeof(buf),MSG_NOSIGNAL);//nacteme
   if (rb==0 || rb==-1) { //EOF od serveru || chyba spojeni
    debug_printf("server EOF\n");
    server_eof=1;
    shutdown(s,1);//uz na klient nelze nic posilat
   } else send_all(s,buf,rb,MSG_NOSIGNAL);//posleme
//   debug_printf("data from server (%d)\n",rb);
  }
//  debug_printf("done\n");
  //zase vrati sety do puvodniho stavu
  FD_ZERO(&fds);
  if (!client_eof) FD_SET(s,&fds);
  if (!server_eof) FD_SET(ss,&fds);
  if (server_eof && client_eof) break;//konec spojeni
 }
fatal_error:
 debug_printf("bailout - end of connection\n");
 buf_free(&b);
 buf_free(&b_remote);
 return ss;
}

int pop3s;//socket na POP3
int smtps;//socket na SMTP

/** zmeni sockety, pokud je po zmene konfigurace nuntno naslouchat na jinych portech */
void main_rebind(int change_smtp,int change_pop3) {
 if (change_smtp) {
  close(smtps);
  smtps=net_listen(active->smtp_port);
 }
 if (change_pop3) {
  close(pop3s);
  pop3s=net_listen(active->pop3_port);
 }
}

/** main - hlavni funkce programu */
int main(int argc,char **argv) {
 int is;
 struct sockaddr_in addr;
 int sof;
 fd_set fds;
 thread_rec tx;

 debug_printf("pop3_proxy\n");

 make_vch_table();
 log_init();
 debug_printf("log_init done\n");
 pthread_mutex_init(&whitelist_lock,NULL);
 config_mutex_lock_init();
 debug_printf("mutex init done\n");
 init_config(argc,argv);
 debug_printf("init_config done\n");

 pop3s=net_listen(active->pop3_port);
 smtps=net_listen(active->smtp_port);
 FD_ZERO(&fds);
 FD_SET(pop3s,&fds);
 FD_SET(smtps,&fds);
 sof=sizeof(addr);
 debug_printf("Awaiting connections: %d %d -> %d\n",smtps,pop3s,max(pop3s,smtps)+1);
 while ((is=select(max(pop3s,smtps)+1,&fds,NULL,NULL,NULL))) {
  if (is==-1 && errno==EINTR) {//signal - nezajima
   debug_printf("signal - ignoring!");
  } else {
   debug_printf("Wake: connection\n");
   if (FD_ISSET(pop3s,&fds)) {
    is=accept(pop3s,(struct sockaddr *)(&addr),&sof);
    tx.func=pop3_handler;
    tx.sock=is;
    debug_printf("pop3\n");
    net_thread_run(tx);
   }
   if (FD_ISSET(smtps,&fds)) {
    is=accept(smtps,(struct sockaddr *)(&addr),&sof);
    tx.func=smtp_handler;
    tx.sock=is;
    debug_printf("smtp\n");
    net_thread_run(tx);
   }
  }
  //zase vrati sety do puvodniho stavu
  FD_ZERO(&fds);
  FD_SET(pop3s,&fds);
  FD_SET(smtps,&fds);
  debug_printf("done\n");
 }
 //sem by se to nemelo dostat
 return EXIT_FAILURE;
}
